from enum import Enum
from pydantic import Field
import random
from functools import reduce
import numpy as np
import torch
import logging
from typing import Any, Callable, List, Tuple, Dict, Optional, Set, Iterable

import torchvision.transforms.functional as F
import torchvision.transforms as transforms
from torchvision.transforms import InterpolationMode
from torchvision.models.detection.transform import resize_boxes
from monai.data.utils import get_random_patch
from monai.transforms.utils import generate_spatial_bounding_box
import albumentations as A

from mmm.BaseModel import BaseModel
from mmm.utils import make_divisable_by
from mmm.data_loading.geojson import GeojsonRegionWindows, GeoAnno
from mmm.logging.type_ext import TransformsSeqType


class TupleToDict:
    def __init__(self, key_names: Optional[Tuple[str, ...]] = None) -> None:
        self.key_names = key_names

    def __call__(self, x: Tuple) -> Dict:
        if self.key_names is not None:
            assert len(x) == len(self.key_names)
            return {k: v for k, v in zip(self.key_names, x)}
        return {str(i): v for i, v in enumerate(x)}


class SaveResultAsKey:
    def __init__(self, key_name: str, f: Callable[[Dict], Any]) -> None:
        self.key_name = key_name
        self.f = f

    def __call__(self, d: Dict) -> Any:
        assert self.key_name not in d, f"{self.key_name} already exists in {d}"
        d[self.key_name] = self.f(d)
        return d


class ApplyToKey:
    def __init__(self, f: Callable, key: str):
        self.f, self.key = f, key

    def __call__(self, d: Dict) -> Dict:
        d[self.key] = self.f(d[self.key])
        return d


class RandomApply:
    def __init__(self, f: Callable, p: float):
        self.f, self.p = f, p

    def __call__(self, d: Dict) -> Dict:
        if random.random() < self.p:
            d = self.f(d)
        return d


class ApplyToKeys:
    """
    >>> dictionary = {"key1": 1, "key2": 2}
    >>> def swap(a,b): return b, a
    >>> from mmm.transforms import ApplyToKeys
    >>> transform = ApplyToKeys(swap, ["key1", "key2"])
    >>> result = transform(dictionary)
    >>> result
    {'key1': 2, 'key2': 1}
    """

    def __init__(self, f: Callable, keys: List[str]):
        self.f, self.keys = f, keys

    def __call__(self, d: Dict) -> Dict:
        # Provide function with arguments corresponding to the dictionary keys
        tuple_result = self.f(*[d[key] for key in self.keys])

        # Play results of function back into dictionary
        for i, key in enumerate(self.keys):
            d[key] = tuple_result[i]

        return d


class ApplyToList:
    """
    Applies a function to a list
    """

    def __init__(self, f: Callable):
        self.f = f

    def __call__(self, ls):
        return [self.f(x) for x in ls]


class ExtractKeysAsTuple:
    def __init__(self, keys: Tuple):
        self.keys = keys

    def __call__(self, d: Dict) -> Tuple:
        return tuple([d[k] for k in self.keys])


class CopyKeysInDict:
    def __init__(self, keys: Dict[str, str]) -> None:
        self.keys = keys

    def __call__(self, d: Dict) -> Dict:
        for key, copy_key in self.keys.items():
            assert copy_key not in d, f"{copy_key} already exists on case {d}"
            d[copy_key] = d[key]
        return d


class KeepOnlyKeysInDict:
    """
    Takes a dictionary and returns that same dictionary but deletes all non-desired elements.

    >>> import numpy as np
    >>> from mmm.transforms import KeepOnlyKeysInDict
    >>> somedictionary = {"keepkey": np.array([1, 1]), "deletethiskey": np.array([2, 2])}
    >>> somedictionary
    {'keepkey': array([1, 1]), 'deletethiskey': array([2, 2])}
    >>> KeepOnlyKeysInDict(keys={"keepkey"})(somedictionary)
    {'keepkey': array([1, 1])}

    If you also want to rename some of the remaining keys, use rename_keys.
    >>> KeepOnlyKeysInDict(keys={"keepkey"}, rename_keys={"keepkey": "newkey"})(somedictionary)
    {'newkey': array([1, 1])}
    """

    def __init__(self, keys: Set[str], rename_keys: Optional[Dict[str, str]] = None) -> None:
        self.keys: Set[str] = keys
        self.rename_dict = rename_keys

    def __call__(self, d: Dict) -> Dict:
        # d changes size during iteration, so copy the iterator with the keys to a list first
        for key in list(d.keys()):
            if key not in self.keys:
                d.pop(key)

        if self.rename_dict is not None:
            for k, v in self.rename_dict.items():
                d[v] = d.pop(k)

        return d


def batchify(ls: List[np.ndarray], warn_when_empty=True):
    """
    Adds an empty batch-dimension to every item in the list and stacks them.
    """
    if ls:
        return np.concatenate([np.expand_dims(x, axis=0) for x in ls], axis=0)
    else:
        if warn_when_empty:
            logging.warning(f"encountered empty object in `batchify`: {ls}")
        return np.array([], np.float32)


class ChannelsSwapFirstLast:
    def __init__(self, keys: Tuple[str]) -> None:
        self.keys = keys

    def __call__(self, d: Dict[str, Any]) -> Any:
        for key in self.keys:
            d[key] = torch.moveaxis(d[key], -1, 0)

        return d


class ResizeWithMask:
    def __init__(self, new_size: Tuple[int, int]) -> None:
        self.img_resizer = transforms.Resize(new_size, interpolation=transforms.InterpolationMode.BILINEAR)
        self.mask_resizer = transforms.Compose(
            [
                lambda x: torch.unsqueeze(x, 0),
                transforms.Resize(new_size, interpolation=transforms.InterpolationMode.NEAREST),
                lambda x: x[0],
            ]
        )

    def __call__(self, image: torch.Tensor, mask: torch.Tensor) -> Any:
        assert len(image.shape) - 1 == len(mask.shape), "Images should have one more dimension than masks"
        return self.img_resizer(image), self.mask_resizer(mask)


class ResizeImage:
    def __init__(self, new_size: List[int], original_size_storage_key: str = "original_size") -> None:
        self.img_resizer = transforms.Resize(new_size, interpolation=transforms.InterpolationMode.BILINEAR)
        self.original_size_storage_key = original_size_storage_key

    def __call__(self, d: Dict[str, Any]) -> Any:
        if self.original_size_storage_key not in d:
            d[self.original_size_storage_key] = list(d["image"].shape[1:])
        else:
            assert (
                len(d["image"].shape) == len(d[self.original_size_storage_key].shape) - 1
            ), "Original shape should only consist of spatial dimensions"

        d["image"] = self.img_resizer(d["image"])
        return d


class ResizeBoxes:
    def __init__(self, overwrite_original_size=None, overwrite_new_size=None) -> None:
        self.overwrite_original_size = overwrite_original_size
        self.overwrite_new_size = overwrite_new_size

    def __call__(self, d: Dict[str, Any]) -> Any:
        if self.overwrite_original_size is None:
            assert "original_size" in d, "Original size of image needs to be known for resizing boxes"
            original_size = d["original_size"]
        else:
            original_size = self.overwrite_original_size

        new_size = d["image"].shape[1:] if self.overwrite_new_size is None else self.overwrite_new_size

        d["boxes"] = resize_boxes(torch.Tensor(d["boxes"]), original_size, new_size)

        return d


class MaskedPatchExtractor:
    """
    Takes in a large image with a mask and outputs a list of patches suitable for `CachingSubCaseDS`.

    Masks should not have a channel dimension: [H, W]

    It exists because MONAI cannot deal well with colorimages.

    To suppress patches with only the background class, it uses a region of interest around the foreground.
    Currently, foreground is defined as any non-zero value.

    Expects channels first images.
    """

    class Config(BaseModel):
        patch_sizes: List[int] = [224]
        sizeaugmentation: float = Field(
            default=0.1,
            description="Jiggle the H, W and coordinates of the patch by a maximum of this factor.",
        )
        max_patches: Optional[int] = None

    def __init__(
        self,
        args: Config,
        patchfilter: Optional[Callable] = None,
        mask_key: str = "label",
    ):
        self.args = args
        self.mask_key = mask_key
        self.patchfilter = patchfilter

        self.min_patch_size = min(args.patch_sizes)
        self.regionextractor = GeojsonRegionWindows(
            coordinate_augmentation="random",
            windowsize_augmentation="relative",
            augmentation_strength=args.sizeaugmentation,
            stepsize=2,
            patch_size=(self.min_patch_size, self.min_patch_size),
        )

    def apply(self, d: Dict) -> Iterable[Dict[str, Any]]:
        img: torch.Tensor = d["image"]
        mask: torch.Tensor = d[self.mask_key]

        rect = GeoAnno.rectangle_builder((0, 0), (mask.shape[-2], mask.shape[-1]))

        pseudo_levels = {i: patchsize / self.min_patch_size for i, patchsize in enumerate(self.args.patch_sizes)}
        g = self.regionextractor.iter_valid_windows(rect, pseudo_levels)
        for i, (level, xy, hw) in enumerate(g):
            hw_scaled = hw[0] * pseudo_levels[level], hw[1] * pseudo_levels[level]
            patchmeta = {"level": level, "xy": xy, "hw": hw_scaled, "i": i}

            x1 = max(0, int(xy[0]))
            x2 = min(mask.shape[-2], int(xy[0] + hw_scaled[0]))
            y1 = max(0, int(xy[1]))
            y2 = min(mask.shape[-1], int(xy[1] + hw_scaled[1]))
            res = {
                "image": img[:, x1:x2, y1:y2],
                self.mask_key: mask[..., x1:x2, y1:y2],
                "meta": {"patchmeta": patchmeta},
            }
            if "meta" in d:
                res["meta"]["supermeta"] = d["meta"]
            if self.patchfilter is None or self.patchfilter(res):
                yield res


def flatten_list(lst: List[List[Any]]) -> List[Any]:
    return [item for sublist in lst for item in sublist]


def _case_to_aformat(
    d: Dict[str, Any]
) -> Tuple[np.ndarray, Optional[np.ndarray], Optional[List[float]], Optional[List[int]]]:
    aimg = (d["image"] * 255.0).permute(1, 2, 0).numpy().astype(np.uint8)
    amask = d["label"].numpy() if "label" in d else None
    amasks = [d["masks"].permute(1, 2, 0).numpy()] if "masks" in d else None
    aboxes = d["boxes"].type(torch.LongTensor).tolist() if "boxes" in d else None
    alabels = d["labels"].tolist() if "labels" in d else None

    if aboxes is not None and alabels is not None:
        invalid_box = [b[0] >= b[2] or b[1] >= b[3] for b in aboxes]
        if True in invalid_box:
            logging.warning(f"Invalid box found in {d['meta']}")
            aboxes = [b for i, b in enumerate(aboxes) if not invalid_box[i]]
            alabels = [l for i, l in enumerate(alabels) if not invalid_box[i]]
    return aimg, amask, amasks, aboxes, alabels


def _atransform_into_case(transformed: Dict[str, Any]) -> Dict[str, Any]:
    res = {"image": F.to_tensor(transformed["image"])}  # torch.from_numpy(transformed['image']).permute(2, 0, 1),
    if "bboxes" in transformed:
        res["boxes"] = torch.Tensor(transformed["bboxes"])
        res["labels"] = torch.Tensor(transformed["class_labels"]).long()
    if "mask" in transformed:
        res["label"] = torch.from_numpy(transformed["mask"]).long()
    if "masks" in transformed:
        res["masks"] = torch.from_numpy(transformed["masks"][0]).permute(2, 0, 1).long()
    return res


class Alb:
    def __init__(self, transforms: TransformsSeqType, support_boxes=False) -> None:
        box_kwargs = (
            {"bbox_params": A.BboxParams(format="pascal_voc", label_fields=["class_labels"])} if support_boxes else {}
        )
        self.transform = A.Compose(
            transforms,
            **box_kwargs,
        )

    def __call__(self, d: Dict[str, Any]) -> Dict[str, Any]:
        aimg, amask, amasks, aboxes, alabels = _case_to_aformat(d)

        transform_kwargs: Dict[str, Any] = dict(image=aimg)
        for kwarg_name, kwarg_value in [
            ("mask", amask),
            ("masks", amasks),
            ("bboxes", aboxes),
            ("class_labels", alabels),
        ]:
            if kwarg_value is not None:
                transform_kwargs[kwarg_name] = kwarg_value
        transformed = self.transform(**transform_kwargs)

        transformed_case = _atransform_into_case(transformed)
        d.update(transformed_case)
        return d


class AlbWithBoxes(Alb):
    def __init__(self, transforms: TransformsSeqType) -> None:
        super().__init__(transforms, support_boxes=True)


class UnifySizes:
    """
    Can be used before a collate_fn in dataloaders to make sure that all images have the same size.

    The size will be divisable by divisable_by to accomodate network constraints.

    The dynamic batch size can be tuned by the max_pixels_in_batch parameter.
    For example, if you think that you can fit a batch size of 96 224x224 images into your GPU memory,
    you can set max_pixels_in_batch to 96*224*224.

    THIS WILL CHANGE THE ORDER INSIDE THE BATCH if not `enforce_order`!
    """

    def __init__(
        self,
        divisable_by=32,
        max_pixels_in_batch=None,
        max_edge_len=None,
        support_boxes=False,
        enforce_order=False,
    ) -> None:
        self.divisable_by = divisable_by
        self.max_pixels_in_batch, self.max_edge_len = max_pixels_in_batch, max_edge_len
        self.support_boxes = support_boxes
        self.enforce_order = enforce_order

    @staticmethod
    def resize_case(d: Dict[str, Any], new_width: int, new_height: int, support_boxes: bool = False) -> Dict[str, Any]:
        return Alb(transforms=[A.Resize(new_height, new_width)], support_boxes=support_boxes)(d)

    @staticmethod
    def get_divisable_dims(img: torch.Tensor, divisable_by: int) -> Tuple[int, int]:
        height = img.shape[1]
        width = img.shape[2]

        width_divisable = make_divisable_by(width, by=divisable_by)
        height_divisable = make_divisable_by(height, by=divisable_by)

        return height_divisable, width_divisable

    def __call__(self, batch: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        # Pick one random element from batch
        sizegiver = random.choice(batch)

        # If the max edge length is not none, first resize the sizegiver keeping its aspect ratio
        if self.max_edge_len is not None:
            if sizegiver["image"].shape[1] > self.max_edge_len or sizegiver["image"].shape[2] > self.max_edge_len:
                # Resizing only needs to happen if the image is larger than the max_edge_len
                sizegiver = Alb(
                    transforms=[A.LongestMaxSize(max_size=self.max_edge_len, always_apply=True)],
                    support_boxes=self.support_boxes,
                )(sizegiver)

        height_divisable, width_divisable = self.get_divisable_dims(sizegiver["image"], self.divisable_by)

        # The batch size might need to be adapted if the image size is changed
        if self.max_pixels_in_batch is not None:
            new_batch_size = self.max_pixels_in_batch // (width_divisable * height_divisable)
            if new_batch_size < len(batch):
                logging.warning(f"UnifySizes had to reduce batch size from {len(batch)} to {new_batch_size}!")
        else:
            new_batch_size = len(batch)

        # Resize all images in the batch to the same size
        filtered_batch = (
            [
                self.resize_case(
                    sizegiver,
                    width_divisable,
                    height_divisable,
                    support_boxes=self.support_boxes,
                )
            ]
            if not self.enforce_order
            else []
        )
        for d in batch:
            # sizegiver is already in the batch
            if (d is sizegiver) and (not self.enforce_order):
                continue

            if len(filtered_batch) < new_batch_size:
                filtered_batch.append(
                    self.resize_case(
                        d,
                        width_divisable,
                        height_divisable,
                        support_boxes=self.support_boxes,
                    )
                )
            else:
                break

        return filtered_batch



import wandb
from typing import Tuple, Literal, Any, Dict
import torch
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
from torchvision.utils import make_grid

from .MTLTask import MTLTask
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.MTLDecoder import MTLDecoder
from mmm.data_loading.MTLDataset import MTLDataset
from mmm.data_loading.TrainValCohort import TrainValCohort

from mmm.logging.type_ext import StepFeedbackDict, StepMetricDict
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules


def _build_criterion(loss_type: Literal["l1", "l2", "bce"]) -> nn.Module:
    if loss_type == "bce":
        return nn.BCELoss()
    if loss_type == "l1":
        return nn.L1Loss()
    if loss_type == "l2":
        return nn.MSELoss()
    else:
        return nn.CrossEntropyLoss()


class ImageGenerationTask(MTLTask):
    """
    Task to be used in pre-training as auxilliary task for representation learning.
    Aims at reconstructing (parts of) the (augmented) input image. Images will not be clear.
    """

    class Config(MTLTask.Config):
        encoder_key: str = "encoder"
        decoder_key: str = "miniDecoder"
        squeezer_key: str = "squeezer"
        reconstruction_loss: Literal["l1", "l2", "bce"] = "l2"
        predicted_channels: int = 3
        scale_factor: float = 30  # MSE is rather low compared to CE, but works better when generating images. To normalize it to other tasks this factor can be used. tested in 224x224 images

    def __init__(
        self,
        args: Config,
        for_decoder: MTLDecoder,
        cohort: TrainValCohort[MTLDataset],
    ) -> None:
        super().__init__(args, cohort)
        self.args: ImageGenerationTask.Config
        self.task_modules: nn.ModuleDict = nn.ModuleDict(
            {
                "head": nn.Sequential(
                    nn.Conv2d(
                        in_channels=for_decoder.get_output_dim_per_pixel(),
                        out_channels=args.predicted_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        bias=False,
                    )
                )
            }
        )

        self.criterion = _build_criterion(self.args.reconstruction_loss)

    def forward(self, x: Any, shared_blocks: Dict[str, SharedBlock]):
        pyr = shared_blocks[self.args.encoder_key](x)

        if hasattr(self.args, "squeezer_key") and self.args.squeezer_key in list(shared_blocks.keys()):
            pyr[-1], _ = shared_blocks[self.args.squeezer_key](pyr)

        decoder_output: torch.Tensor = shared_blocks[self.args.decoder_key](pyr)

        return self.task_modules["head"](decoder_output)

    def prepare_batch(self, batch: Dict[str, Any]) -> Any:
        batch["image"] = batch["image"].to(self.torch_device)
        batch["target"] = batch["target"].to(self.torch_device)

        return batch

    def training_step(
        self, batch: Dict[str, Any], shared_modules: SharedModules
    ) -> Tuple[torch.Tensor, StepFeedbackDict]:
        input_batch, target_batch = batch["image"], batch["target"]
        pred_batch = shared_modules.forward(input_batch, self.forward)

        # scale to prediction output size
        size = pred_batch.shape[-2:]
        target_batch = F.interpolate(target_batch, size=size, mode="bicubic")

        # for vis
        input_batch = F.interpolate(input_batch, size=size, mode="bicubic")

        loss = self.criterion(torch.sigmoid(pred_batch), torch.sigmoid(target_batch)) * self.args.scale_factor
        step_results: StepMetricDict = {}

        self.add_step_result(loss.item(), step_results)
        live_vis = self._visualize_step(
            source_img=input_batch.detach().cpu()[0],
            target_img=torch.sigmoid(target_batch).cpu()[0],
            pred_img=torch.sigmoid(pred_batch.detach().cpu())[0],
        )
        return loss, live_vis

    def _visualize_step(self, source_img, target_img, pred_img):
        vis_n = min(self._takeout_vis_budget(), source_img.size(0))
        if vis_n <= 0:
            return {}
        shape = pred_img.shape[-2:]
        target_img = F.interpolate(target_img.unsqueeze(0), size=shape).squeeze()
        pred_img = F.interpolate(pred_img.unsqueeze(0), size=shape).squeeze()
        diffimg = torch.abs(target_img - pred_img)
        strs = [
            f"Source: {np.min(source_img.numpy()):.3f}, {np.max(source_img.numpy()):.3f}, shape:{source_img.shape}\n"
            f"Target: {np.min(target_img.numpy()):.3f}, {np.max(target_img.numpy()):.3f}, shape:{target_img.shape}\n"
            f"Pred: {np.min(pred_img.numpy()):.3f}, {np.max(pred_img.numpy()):.3f}, shape:{pred_img.shape}\n"
            f"Difference: {torch.sum(diffimg)}"
        ]

        return {
            "vis-channel-split": wandb.Image(
                make_grid(
                    [
                        source_img[0].unsqueeze(0),
                        target_img[0].unsqueeze(0),
                        pred_img[0].unsqueeze(0),
                        source_img[1].unsqueeze(0),
                        target_img[1].unsqueeze(0),
                        pred_img[1].unsqueeze(0),
                        source_img[2].unsqueeze(0),
                        target_img[2].unsqueeze(0),
                        pred_img[2].unsqueeze(0),
                    ],
                    nrow=9,
                ),
                caption="\n".join(strs),
            ),
            "vis-img": wandb.Image(
                make_grid([source_img, target_img, pred_img], nrow=3),
                caption="\n".join(strs),
            ),
        }



from typing import Dict, Any, Sequence, Union
import numpy as np
import albumentations as A

# A step metric dict can be aggregated by concatenating the arrays for each key
# using `utils.flatten_list_of_dicts`
StepMetricDict = Dict[str, np.ndarray]
# A step feedback dict can be processed by trainers to give training feedback during a loop
StepFeedbackDict = Dict[str, Any]

TransformsSeqType = Sequence[Union[A.BasicTransform, A.BaseCompose]]



from typing import Literal, List, Optional
from pydantic import Field
import torch
import torch.nn as nn
import torchvision.models as torch_models
from torchvision.models.convnext import (
    ConvNeXt,
    ConvNeXt_Tiny_Weights,
    LayerNorm2d,
    ConvNeXt_Small_Weights,
)

from mmm.torch_ext import infer_stride_channels_from_features
from mmm.neural.TorchModule import TorchModule
from mmm.mtl_modules.shared_blocks import SharedBlock
from mmm.neural.activations import ActivationFunctionConfig, ActivationFn
from mmm.torch_ext import replace_childen_recursive


def replace_with_defaultnorm(old_norm: LayerNorm2d):
    return nn.BatchNorm2d(num_features=old_norm.normalized_shape[0], eps=old_norm.eps)


WEIGHTSKB = {
    "tiny": (torch_models.convnext_tiny, ConvNeXt_Tiny_Weights.DEFAULT),
    "small": (torch_models.convnext_small, ConvNeXt_Small_Weights.DEFAULT),
}


class TorchVisionConvnext(nn.Module):
    """
    For MTL, replace all layernorms with layernorms without learnable parameters!
    """

    class Config(TorchModule):
        architecture: Literal["convnext"] = "convnext"
        pretrained: bool = True
        # use_latent_layer: bool = False
        # latent_activation: ActivationFunctionConfig = Field(
        #     default=ActivationFunctionConfig(fn_type=ActivationFn.ReLU),
        #     description="Only relevant when used with `use_latent_layer`."
        # )
        variant: Literal["tiny", "small"] = "tiny"

        def build_instance(self, hidden_dim: int) -> nn.Module:
            # Avoid cyclical import by importing here
            return TorchVisionConvnext(self, hidden_dim)

    def __init__(self, args: Config, hidden_dim: int) -> None:
        super().__init__()
        self.args = args

        if self.args.pretrained:
            self.wrapped_model: ConvNeXt = WEIGHTSKB[self.args.variant][0](weights=WEIGHTSKB[self.args.variant][1])
        else:
            self.wrapped_model: ConvNeXt = WEIGHTSKB[self.args.variant][0]()

        # Sharedblocks will later replace the default norm with a good norm
        replace_childen_recursive(self.wrapped_model, LayerNorm2d, replace_with_defaultnorm)

        # Delete the last fc layer, it is not required
        self.wrapped_model.classifier = self.wrapped_model.classifier[:2]

        with torch.no_grad():
            self.channels, self.strides = infer_stride_channels_from_features(
                self._feature_forward(torch.rand(1, 3, 224, 224))
            )
        # assert hidden_dim == self.get_feature_pyramid_channels()[-1]

    def _feature_forward(self, x: torch.Tensor):
        """
        Given the original input, returns a list with the input and all feature maps
        """
        activations = [x]
        for layer in self.wrapped_model.features:
            activations.append(layer(activations[-1]))

        return activations[::2]

    def forward(self, x: torch.Tensor):
        features = self._feature_forward(x)
        # Apply avg pool
        z = self.wrapped_model.avgpool(features[-1])
        z = torch.flatten(z, 1)
        return features, z

    def get_strides(self):
        # Input and spatial sizes of the feature maps
        # return [1] + [4, 8, 16, 32]
        return self.strides

    def get_feature_pyramid_channels(self) -> List[int]:
        # [3, 96, 192, 384, 768]
        return self.channels



from pydantic_settings import BaseSettings
from pathlib import Path


class Settings(BaseSettings):
    max_classes_detailed_logging: int = 25
    ignore_class_value: int = -1

    class Config:
        env_prefix = "mtl_"


mtl_settings = Settings()



from typing import Any, List, Optional, Callable, Dict

import numpy as np
import torch
from torch.utils.data import Dataset

from mmm.logging.st_ext import blend_with_mask
from .MTLDataset import MTLDataset, SrcCaseType


class SemSegDataset(MTLDataset):
    """
    Contains a single pixel-dense mask for one or more images.
    """

    def __init__(
        self,
        src_ds: Dataset[SrcCaseType],
        class_names: List[str] = None,
        *args,
        **kwargs,
    ) -> None:
        assert class_names is not None, "Segmentation dataset needs class names"
        self.class_names: List[str] = class_names
        super().__init__(src_ds, ["image", "label"], ["original_shape", "meta"], *args, **kwargs)

    def set_classes_for_visualization(self, classes: List[str]):
        self.class_names = classes

    def verify_case(self, d: Dict[str, Any]) -> None:
        super().verify_case(d)
        self.assert_image_data_assumptions(d["image"])
        # For example, MONAI tends to load some specific subclass of Tensor, exclude it!
        assert torch.is_tensor(d["label"]), "mask should be of type tensor"
        assert isinstance(d["label"], torch.LongTensor)
        assert d["label"].shape == d["image"].shape[1:], f"{d['label'].shape=} should match {d['image'].shape[1:]=}"
        assert torch.max(d["label"]).item() < len(self.class_names), "There need to be more class names"
        assert len(d["image"].shape) == (len(d["label"].shape) + 1), "all labels should be in a one-dim tensor"

    def st_case_viewer(self, case: Dict[str, Any], i: int = -1) -> None:
        import streamlit as st

        st.title("Untransformed image:")
        blend_with_mask(
            case["image"],
            case["label"],
            classes=self.class_names,
            st_key=f"c{i}",
        )
        st.write({k: v for k, v in case.items() if k not in ["image", "label"]})

    def _compute_batchsize_from_batch(self, batch: Dict[str, Any]) -> int:
        return batch["image"].shape[0]

    def _visualize_batch_case(self, batch: Dict[str, Any], i: int) -> None:
        import streamlit as st

        patch = batch["image"][i]
        patch_mask = batch["label"][i]
        if "meta" in batch:
            st.write(batch["meta"][i])

        mask_uniques = torch.unique(patch_mask)
        blend_with_mask(
            patch,
            patch_mask,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}\nUniques: {mask_uniques}",
            classes=self.class_names,
            st_key=f"b{i}",
        )



from __future__ import annotations
from typing import Any, List, Dict, Mapping, Tuple, Optional, Type, Literal
import random
from pydantic import Field

import wandb
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset

from mmm.logging.type_ext import StepMetricDict

from .MTLTask import MTLTask
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.data_loading.MultilabelClassificationDataset import (
    MultilabelClassificationDataset,
)
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.logging.wandb_ext import remove_wandb_special_chars
from mmm.neural.pooling import GlobalPooling, GlobalPoolingConfig

from mmm.utils import flatten_list_of_dicts
from sklearn.metrics import roc_auc_score, accuracy_score, f1_score


class MultilabelClassificationTask(MTLTask):
    """
    Multilabel task. Takes a single image as input and can predict multiple binary classes.
    If you have multiple multi-class problems, you either need to convert that problem
    into a binary problem for each class or multiple multi-class classification tasks.

    Loss weights are optional.
    If a loss weight is zero this class will not contribute to the loss or metrics.

    For AUC computation, classes completely missing are ignored.
    """

    class Config(MTLTask.Config):
        encoder_key: str = "encoder"
        squeezer_key: str = "squeezer"
        confidence_threshold: float = 0.5
        dropout: float = 0.2
        metrics: List[Literal["acc", "f1", "auc"]] = ["acc", "auc"]

    def __init__(
        self,
        hidden_dim: int,
        args: Config,
        cohort: TrainValCohort[MultilabelClassificationDataset],
    ):
        super().__init__(args, cohort)
        self.args: MultilabelClassificationTask.Config
        self.hidden_dim: int = hidden_dim
        self.class_names: List[str] = cohort.datasets[0].class_names

        self.flatten = nn.Flatten(1)
        self.task_modules = nn.ModuleDict(
            {
                "classification_head": nn.Sequential(
                    nn.Dropout(p=self.args.dropout),
                    nn.ReLU(),
                    nn.Linear(self.hidden_dim, len(self.class_names)),
                )
            }
        )

    def prepare_batch(self, batch: Dict[str, Any]) -> Any:
        batch["image"] = batch["image"].to(self.torch_device)
        batch["class_labels"] = batch["class_labels"].to(self.torch_device)
        if "loss_weights" in batch:
            batch["loss_weights"] = batch["loss_weights"].to(self.torch_device)
        return batch

    def forward(self, x, shared_blocks: Dict[str, SharedBlock]):
        # assert shared_blocks[self.args.encoder_key].training == self.training,\
        #     f"Encoder is in different state than {self.get_name()}"
        pyr = shared_blocks[self.args.encoder_key](x)
        _, squeezed = shared_blocks["squeezer"](pyr)
        hidden_vector = self.flatten(squeezed)
        out = self.task_modules["classification_head"](hidden_vector)
        return out

    def training_step(self, batch: Dict[str, Any], shared_blocks: SharedModules) -> Tuple[torch.Tensor, Dict]:
        x = batch["image"]
        y = batch["class_labels"]
        w = batch["loss_weights"] if "loss_weights" in batch else None

        y_hat = shared_blocks.forward(x, self.forward)

        batch_loss = F.binary_cross_entropy_with_logits(y_hat, y, weight=w) * 1.44269  # log2(exp(1))
        if w is not None:
            batch_loss = batch_loss / w.mean()

        with torch.no_grad():
            step_results: StepMetricDict = {  # type: ignore (.numpy() does not correctly indicate numpy array)
                "targets": y.cpu().numpy(),
                "logits": y_hat.detach().cpu().numpy(),
                "preds": torch.sigmoid(y_hat).detach().cpu().numpy(),
                "weights": (w.detach().cpu().numpy() if w is not None else np.ones(y.shape)),
            }
            self.add_step_result(batch_loss.item(), step_results)

            live_vis = self._visualize_preds(
                x.detach().cpu(),
                step_results,
                batch["meta"] if "meta" in batch else None,
            )

        return batch_loss, live_vis

    def _visualize_preds(self, training_ims, step_metrics: Dict, metainfo: Optional[List[Dict]]) -> Dict[str, Any]:
        vis_n = min(self._takeout_vis_budget(), training_ims.size(0))

        if vis_n <= 0:
            return {}

        preds = []
        for rand_index in random.sample(list(range(training_ims.size(0))), vis_n):
            predstrs = [
                f"{cls_name}({target})>{pred:.2f}"
                for cls_name, target, pred in zip(
                    self.class_names,
                    step_metrics["targets"][rand_index],
                    step_metrics["preds"][rand_index],
                )
            ]
            preds.append(wandb.Image(training_ims[rand_index], caption="\n".join(predstrs)))
        return {"preds": preds} if preds else {}

    def _get_short_class_names(self, max_length=10):
        if True in [len(c) > max_length for c in self.class_names]:
            return [f"{i};{c[:max_length]}" for i, c in enumerate(self.class_names)]
        else:
            return self.class_names

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        metric_computers = {"acc": accuracy_score, "f1": f1_score}
        metrics = flatten_list_of_dicts(self._step_metrics)

        _, print_str = super().log_epoch_metrics()
        log_dict = {}

        preds_per_class = [
            metrics["preds"][:, i] > self.args.confidence_threshold for i, _ in enumerate(self.class_names)
        ]
        weights_per_class = [metrics["weights"][:, i] for i, _ in enumerate(self.class_names)]

        for metric_literal in self.args.metrics:
            if metric_literal in ["acc", "f1"]:
                metrics_per_class = {
                    f"{self.class_names[i]}_{metric_literal}": metric_computers[metric_literal](
                        y_true=metrics["targets"][:, i],
                        y_pred=preds_per_class[i],
                        sample_weight=weights_per_class[i],
                    )
                    for i, _ in enumerate(self.class_names)
                }
            elif metric_literal == "auc":
                metrics_per_class = {
                    f"{self.class_names[i]}_{metric_literal}": roc_auc_score(
                        y_true=metrics["targets"][:, i],
                        y_score=metrics["logits"][:, i],
                        sample_weight=weights_per_class[i],
                    )
                    for i, _ in enumerate(self.class_names)
                    if len(np.unique(metrics["targets"][:, i])) > 1
                }
                metrics_per_class = {k: v for k, v in metrics_per_class.items() if not np.isnan(v)}
            else:
                raise Exception(f"Unknown metric {metric_literal}")
            log_dict[f"{metric_literal}_mean"] = np.mean(list(metrics_per_class.values()))  # type: ignore
            print_str = f"{print_str} - {metric_literal}_mean: {log_dict[f'{metric_literal}_mean']}"
            for class_name, v in metrics_per_class.items():
                log_dict[f"{remove_wandb_special_chars(class_name)}"] = v

        return log_dict, print_str



from mmm.data_loading.MTLDataset import MTLDataset, DatasetStyle
from mmm.data_loading.TrainValCohort import TrainValCohort, DataSplit
from mmm.data_loading.ClassificationDataset import ClassificationDataset
from mmm.data_loading.MultilabelClassificationDataset import (
    MultilabelClassificationDataset,
)
from mmm.data_loading.RegressionDataset import RegressionDataset
from mmm.data_loading.ImageGenerationDataset import ImageGenerationDataset
from mmm.data_loading.SemSegDataset import SemSegDataset
from mmm.data_loading.MultiLabelSemSegDataset import MultiLabelSemSegDataset
from mmm.data_loading.DetectionDataset import DetectionDataset
from mmm.data_loading.MultipleInstanceDataset import MultipleInstanceDataset
from mmm.data_loading.ImageCaptionDataset import ImageCaptionDataset
from mmm.data_loading.ssl import get_combined_SSL_dataset
from mmm.data_loading.DiskDataset import DiskDataset
from mmm.data_loading.DistributedPath import DistributedPath

from mmm.data_loading.utils import (
    train_val_split,
    train_val_split_class_dependent,
    TransformedSubset,
)
from mmm.data_loading.utils import batch_concat, batch_concat_with_meta
from mmm.utils import (
    load_config_from_env,
    load_config_from_str,
    load_config_from_json5,
)
from mmm.torch_ext import SubCaseDataset, CachingSubCaseDS



from __future__ import annotations
import logging
import numpy as np
import random
import warnings
from typing import Any, List, Optional, Callable, Dict, Tuple, TypeVar

from PIL.Image import Image

import torchvision.transforms.functional as F

from torch.utils.data import Dataset

from mmm.bucketizing import BucketConfig
from mmm.logging.st_ext import blend_with_mask
from .MTLDataset import MTLDataset, SrcCaseType

CaseDict = TypeVar("CaseDict", bound=Dict)


class ClassificationDataset(MTLDataset):
    @staticmethod
    def bucketize_case(
        buckets: BucketConfig, description_extractor: Callable[[CaseDict], str]
    ) -> Callable[[CaseDict], CaseDict]:
        def f(casedict: CaseDict):
            desc = description_extractor(casedict)
            bucket_name, new_class_id = buckets.get_bucket_name(desc)

            assert "bucket_name" not in casedict
            casedict["bucket_name"] = bucket_name

            casedict["class"] = new_class_id
            # if "class"
            # assert "old_class_id" not in casedict
            # casedict["old_class_id"] = casedict["class"]

            return casedict

        return f

    @staticmethod
    def TorchvisionToMTLClf(t: Tuple[Image, int]):
        return {"image": F.to_tensor(t[0]), "class": t[1]}

    @staticmethod
    def from_torchvision(ds: Dataset, class_names: Optional[List[str]] = None) -> ClassificationDataset:
        """
        A torchvision dataset returns tuples of the form (img: PIL.Image, class_index: int)
        """
        warnings.warn("Use TorchvisionToMTLClf instead as a src_transform", DeprecationWarning)
        r = ClassificationDataset(
            src_ds=ds,
            src_transform=lambda t: {"image": F.to_tensor(t[0]), "class": t[1]},
            class_names=class_names,
        )
        return r

    def __init__(
        self,
        src_ds: Dataset[SrcCaseType],
        class_names: Optional[List[str]] = None,
        *args,
        **kwargs,
    ) -> None:
        assert class_names is not None, "Classification datasets without class names are deprecated"
        self.class_names: List[str] = class_names
        self.vis_classes = class_names
        super().__init__(src_ds, ["image", "class"], ["meta"], *args, **kwargs)

    def verify_case_by_index(self, index: int) -> Dict[str, Any]:
        case = super().verify_case_by_index(index)
        self.assert_image_data_assumptions(case["image"])
        assert isinstance(case["class"], int), "Class label should be an integer"
        return case

    def set_indices_by_fraction(self, fraction: float, seed: int = 13) -> None:
        """
        Starts by adding one example per class, then draws random samples until the fraction criterion is met.

        As a result, a fraction of zero will result in a subset of one sample per class
        """
        # Can be used for sampling new cases when artifically reducing the dataset's size
        self.seeded_random = random.Random(seed)
        src_dataset_length = len(self.src_ds)  # type: ignore
        # self._indices = list(range(src_dataset_length))
        self.reset_indices()
        shuffled_original_indices = list(self._indices)
        self.seeded_random.shuffle(shuffled_original_indices)
        new_indices = []

        # Add one sample per class
        classes: List[int] = []
        for original_index in shuffled_original_indices:
            case = self.verify_case_by_index(original_index)
            if (case_class := case["class"]) not in classes:
                classes.append(case_class)
                new_indices.append(original_index)

            if len(new_indices) >= len(self.class_names):
                break
        assert len(new_indices) == len(self.class_names)

        number_of_cases_missing = int(src_dataset_length * fraction) - len(new_indices)
        if number_of_cases_missing > 0:
            unused_indices = set(range(src_dataset_length)) - set(new_indices)
            unbalanced_indices = self.seeded_random.sample(unused_indices, number_of_cases_missing)
            new_indices.extend(unbalanced_indices)
        self._indices = np.array(new_indices)

    def set_classes_for_visualization(self, classes: List[str]):
        self.class_names = classes

    def get_classes_for_visualization(self):
        return self.class_names

    def get_input_output_tuple(self, batch: Dict[str, Any]) -> Tuple[Any, ...]:
        return batch["image"], batch["class"]

    def st_case_viewer(self, case: Dict[str, Any], i: int) -> None:
        import streamlit as st

        if self.class_names:
            for i, class_name in enumerate(self.class_names):
                if case["class"] != i:
                    # For large datasets, do not show negative classes
                    if len(self.class_names) < 100:
                        st.write(f"class {i}: {class_name}")
                else:
                    st.write(f"THIS CLASS ({i}): {class_name}")

        im = case["image"]
        blend_with_mask(im, None, caption_suffix=f"Shape: {im.shape}", st_key=f"c{i}")
        st.write(case)

    def _compute_batchsize_from_batch(self, batch: Dict[str, Any]) -> int:
        return batch["image"].shape[0]

    def _visualize_batch_case(self, batch: Dict[str, Any], i: int) -> None:
        import streamlit as st

        patch = batch["image"][i]
        class_name = self.class_names[batch["class"][i]]
        st.write(f"Label: {batch['class'][i]}, " + class_name)

        try:
            st.write(batch["meta"][i])
        except:
            logging.debug(f"Batch does not contain meta information at {i}")

        blend_with_mask(
            patch,
            None,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}: {patch.shape}",
            st_key=f"b{i}",
        )



"""
Geojson is a common format for storing the annotations on gigapixel 2D images.

Our GeoJSON utility helps with two use-cases:

## Region annotations

Extracting windows from annotated regions, often used in multiclass classification.
A single annotations is usually larger than a window.

## Object annotations
 
Extracting windows with annotated objects, often used in detection and segmentation.
A single annotation is usually smaller than a window and enables multi-level window extraction.

"""

from .GeoAnno import GeoAnno, AnnotationType
from .AnnoFilterConfig import AnnoFilterConfig
from .GeojsonRegionWindows import GeojsonRegionWindows
from .GeojsonObjWindows import GeojsonObjWindows
from .NoUsefulWindowException import NoUsefulWindowException
from .WSIGeojsonDataset import WSIGeojsonDataset



from __future__ import annotations
from pydantic import Field
from typing import Any, List, Dict, Tuple, Optional, Mapping
from abc import abstractmethod

import torch
import torch.nn as nn

from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.data_loading.MTLDataset import MTLDataset

from ..MTLModule import MTLModule
from mmm.mtl_modules.MTLModule import MTLModule
from mmm.event_selectors import EventSelector, RecurringEventSelector
from mmm.logging.type_ext import StepFeedbackDict, StepMetricDict
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules


class MTLTask(MTLModule):
    """
    Task which makes use of shared blocks during its forward pass.

    Guidelines when developing your own task types:

    Implementations need to extend/override abstractmethods such as `forward` and `training_step`.

    Giving feedback can happen after each step using the `training_step` implementation
    and after each loop/epoch/... by overwriting `log_epoch_metrics`.

    Loss is expected to be approx. 1 given random input. This makes sure your task type is not dominated nor dominates
    the framework's task types.
    """

    class Config(MTLModule.Config):
        vis_enabled_epochs: EventSelector = Field(
            default=RecurringEventSelector(starting_at=0, every_n=1),
            description="Controls the loops with enabled visualization.",
        )
        max_visualizations_per_full_train_loop: int = 2
        max_visualizations_per_full_val_loop: int = 4

    def __init__(self, args: Config, cohort: TrainValCohort[MTLDataset]) -> None:
        super().__init__(args)
        self.args: MTLTask.Config = args
        self.cohort: TrainValCohort[MTLDataset] = cohort
        self.task_modules: nn.ModuleDict

        self._step_losses: List[float] = []
        self._step_metrics: List[StepMetricDict] = []

        self._vis_number_buffer: float = 0.0

    def _takeout_vis_budget(self) -> int:
        """
        Returns the task's visualization budget at the current step.

        Relies on the module's `training: bool` property.

        Tries to induce the total number of steps from the dataloader's length if the dataset is map-style.
        """
        # Adjust the vis_enabled_epochs to enable visualizations for an epoch
        if not self.args.vis_enabled_epochs.is_event(self._epoch):
            return 0

        # current_step = len(self._step_losses)
        # dataloader_length = len(self.cohort.data_loaders[0 if self.training else 1])  # type: ignore
        if self.training:
            max_vis = self.args.max_visualizations_per_full_train_loop
        else:
            max_vis = self.args.max_visualizations_per_full_val_loop

        self._vis_number_buffer += 1
        return 1 if self._vis_number_buffer < max_vis else 0

    def prepare_batch(self, batch: Dict[str, Any]) -> Any:
        """
        Prepares the batch for training on the correct device.

        The trainer calls this funci

        By default, it will be attempted to copy the whole data structure to GPU
        """
        raise NotImplementedError

    def add_step_result(self, loss: float, step_metrics: StepMetricDict):
        self._step_losses.append(loss)
        self._step_metrics.append(step_metrics)

    def get_step_losses(self):
        return self._step_losses

    def is_currently_being_trained(self) -> bool:
        according_to_loss = len(self._step_losses) != 0
        according_to_metrics = len(self._step_metrics) != 0
        assert according_to_loss == according_to_metrics, f"Missing logging of metrics for MTLModule {self.get_name()}?"
        return according_to_loss

    def get_short_status(self) -> str:
        if self.is_currently_being_trained():
            avg_loss = sum(self.get_step_losses()) / len(self.get_step_losses())
            return f"{avg_loss:.3f}"
        else:
            return "-"

    @abstractmethod
    def forward(self, x: Any, shared_blocks: SharedModules):
        raise NotImplementedError

    @abstractmethod
    def training_step(
        self, batch: Dict[str, Any], shared_blocks: SharedModules
    ) -> Tuple[torch.Tensor, StepFeedbackDict]:
        """
        Returns the loss (single number as torch.FloatTensor) and a dictionary with immediate feedback for a human.

        Here, a task should store metrics for later aggregation using `self.add_step_result(...)`.
        """
        raise NotImplementedError()

    def reset_metric_buffers(self):
        self._step_losses = []
        self._step_metrics = []

    def prepare_epoch(self, epoch: int, prefix: str, training_mode: Optional[bool] = None):
        """
        Called before starting an epoch. Used to reset buffer and prepare dataloading for cross validation.
        """
        super().prepare_epoch(epoch, prefix, training_mode=training_mode)
        self.reset_metric_buffers()
        self._vis_number_buffer = 0

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        """
        Called when the user of the task wants the task to log its metric buffers.

        Happens at the end of a loop and can be an expensive, infrequent operation.

        .. code-block::
            :caption: Example dictionary from the metrics buffer

            {
                "acc": 0.96,
                "auc": 0.76
            }

        The string is used for printing and should be one line.
        """
        return {}, f"{self.get_name()} - loss: {self.get_short_status()}"

    def __repr__(self) -> str:
        res = super().__repr__()
        res = f"{res}\nconfig={self.args.model_dump_json(indent=2)}\n{self.cohort}"
        return res

    def __repr_html__(self) -> str:
        return f"""{super().__repr_html__()}
        <br />
        {self.cohort.__repr_html__()}
        """
